#!/usr/bin/python
# -*- coding: utf-8 -*-

import inspect
import logging
import os
import sys
import time
import unittest
import urllib.error
import urllib.parse
import urllib.request
from urllib.parse import parse_qs, parse_qsl, urlparse
from urllib.request import Request

logging.basicConfig()

# prefer local copy to the one which is installed
# hack from http://stackoverflow.com/a/6098238/280539
_top_level_path = os.path.realpath(
    os.path.abspath(
        os.path.join(os.path.split(inspect.getfile(inspect.currentframe()))[0], "..")
    )
)
if _top_level_path not in sys.path:
    sys.path.insert(0, _top_level_path)
# end of hack

import warnings
from io import StringIO

# we don't want to let Wrapper do real web-requests. so, we are…
# constructing a simple Mock!
from urllib.error import HTTPError

warnings.simplefilter("always")

import SPARQLWrapper.Wrapper as _victim
from SPARQLWrapper import (
    BASIC,
    CSV,
    DIGEST,
    GET,
    INSERT,
    JSON,
    JSONLD,
    N3,
    POST,
    POSTDIRECTLY,
    RDF,
    RDFXML,
    SELECT,
    TSV,
    TURTLE,
    URLENCODED,
    XML,
    SPARQLWrapper,
)
from SPARQLWrapper.Wrapper import (
    EndPointInternalError,
    EndPointNotFound,
    QueryBadFormed,
    QueryResult,
    Unauthorized,
    URITooLong,
)


class FakeResult(object):
    def __init__(self, request):
        self.request = request


def urlopener(request):
    return FakeResult(request)


def urlopener_error_generator(code):
    def urlopener_error(request):
        raise HTTPError(request.get_full_url, code, "", {}, StringIO(""))

    return urlopener_error


def urlopener_check_data_encoding(request):
    if isinstance(request.data, str):
        raise TypeError


# DONE


class SPARQLWrapper_Test(unittest.TestCase):
    @staticmethod
    def _get_request(wrapper):
        return wrapper.query().response.request  # possible due to mock above

    @staticmethod
    def _get_parameters_from_request(request):
        if request.get_method() == "GET":
            pieces_str = urlparse(request.get_full_url()).query
        else:
            pieces_str = request.data.decode("ascii")

        return parse_qs(pieces_str)

    @staticmethod
    def _get_request_parameters(wrapper):
        request = SPARQLWrapper_Test._get_request(wrapper)
        parameters = SPARQLWrapper_Test._get_parameters_from_request(request)

        return parameters

    @staticmethod
    def _get_request_parameters_as_bytes(wrapper):
        request = SPARQLWrapper_Test._get_request(wrapper)
        parameters = SPARQLWrapper_Test._get_parameters_from_request(request)

        result = {}
        for k, vs in parameters.items():
            result[k] = [v.encode("utf-8") for v in vs]
        return result

    @classmethod
    def setUpClass(cls):
        urllib.request._opener = None  # clear value. Due to the order of test execution, the value of urllib.request._opener contains, for instance, keepalive.keepalive.HTTPHandler

    def setUp(self):
        self.wrapper = SPARQLWrapper(endpoint="http://example.org/sparql")
        _victim.urlopener = urlopener

    def testConstructor(self):
        try:
            SPARQLWrapper()
            self.fail("SPARQLWrapper constructor should fail without arguments")
        except TypeError:
            pass

        wrapper = SPARQLWrapper(endpoint="http://example.org/sparql/")

        self.assertEqual(XML, wrapper.returnFormat, "default return format is XML")
        self.assertTrue(
            wrapper.agent.startswith("sparqlwrapper"),
            'default user-agent should start with "sparqlwrapper"',
        )

        wrapper = SPARQLWrapper(
            endpoint="http://example.org/sparql/", returnFormat="wrongformat"
        )
        self.assertEqual(XML, wrapper.returnFormat, "default return format is XML")

        wrapper = SPARQLWrapper(
            endpoint="http://example.org/sparql/",
            defaultGraph="http://example.org/default",
        )
        parameters = self._get_request_parameters(wrapper)
        self.assertEqual(
            ["http://example.org/default"],
            parameters.get("default-graph-uri"),
            "default graph is set",
        )

    def testReset(self):
        self.wrapper.setMethod(POST)
        self.wrapper.setQuery("CONSTRUCT WHERE {?a ?b ?c}")
        self.wrapper.setReturnFormat(N3)
        self.wrapper.addParameter("a", "b")
        self.wrapper.setOnlyConneg(True)

        request = self._get_request(self.wrapper)
        parameters = self._get_parameters_from_request(request)
        onlyConneg = self.wrapper.onlyConneg

        self.assertEqual("POST", request.get_method())
        self.assertTrue(parameters["query"][0].startswith("CONSTRUCT"))
        self.assertTrue("rdf+n3" in request.get_header("Accept"))
        self.assertTrue("a" in parameters)
        self.assertTrue(onlyConneg)

        self.wrapper.resetQuery()

        request = self._get_request(self.wrapper)
        parameters = self._get_parameters_from_request(request)
        onlyConneg = self.wrapper.onlyConneg

        self.assertEqual("GET", request.get_method())
        self.assertTrue(parameters["query"][0].startswith("SELECT"))
        self.assertFalse("rdf+n3" in request.get_header("Accept"))
        self.assertTrue("sparql-results+xml" in request.get_header("Accept"))
        self.assertFalse("a" in parameters)
        self.assertFalse("a" in parameters)
        self.assertTrue(onlyConneg)

    def testSetReturnFormat(self):
        with warnings.catch_warnings(record=True) as w:
            self.wrapper.setReturnFormat("nonexistent format")
            # Ignore format 'nonexistent format'; current instance supports: ...
            self.assertEqual(1, len(w), "Warning due to non expected format")

        self.assertEqual(XML, self.wrapper.query().requestedFormat)

        self.wrapper.setReturnFormat(JSON)
        self.assertEqual(JSON, self.wrapper.query().requestedFormat)

        self.wrapper.setReturnFormat(JSONLD)
        self.assertEqual(JSONLD, self.wrapper.query().requestedFormat)

    def testsSupportsReturnFormat(self):
        self.assertTrue(self.wrapper.supportsReturnFormat(XML))
        self.assertTrue(self.wrapper.supportsReturnFormat(JSON))
        self.assertTrue(self.wrapper.supportsReturnFormat(TURTLE))
        self.assertTrue(self.wrapper.supportsReturnFormat(N3))
        self.assertTrue(self.wrapper.supportsReturnFormat(RDF))
        self.assertTrue(self.wrapper.supportsReturnFormat(RDFXML))
        self.assertTrue(self.wrapper.supportsReturnFormat(CSV))
        self.assertTrue(self.wrapper.supportsReturnFormat(TSV))
        self.assertFalse(self.wrapper.supportsReturnFormat("nonexistent format"))

        self.assertTrue(self.wrapper.supportsReturnFormat(JSONLD))

    def testAddParameter(self):
        self.assertFalse(self.wrapper.addParameter("query", "dummy"))
        self.assertTrue(self.wrapper.addParameter("param1", "value1"))
        self.assertTrue(self.wrapper.addParameter("param1", "value2"))
        self.assertTrue(self.wrapper.addParameter("param2", "value2"))

        pieces = self._get_request_parameters(self.wrapper)

        self.assertTrue("param1" in pieces)
        self.assertEqual(["value1", "value2"], pieces["param1"])
        self.assertTrue("param2" in pieces)
        self.assertEqual(["value2"], pieces["param2"])
        self.assertNotEqual(["dummy"], "query")

    def testSetCredentials(self):
        request = self._get_request(self.wrapper)
        self.assertFalse(request.has_header("Authorization"))

        self.wrapper.setCredentials("login", "password")
        request = self._get_request(self.wrapper)
        self.assertTrue(request.has_header("Authorization"))

        # expected header for login:password
        # should succeed for python 3 since pull request #72
        self.assertEqual(
            "Basic bG9naW46cGFzc3dvcmQ=", request.get_header("Authorization")
        )

    def testAddCustomHttpHeader(self):
        request = self._get_request(self.wrapper)
        self.assertFalse(request.has_header("Foo"))

        # Add new header field name
        self.wrapper.addCustomHttpHeader("Foo", "bar")
        request = self._get_request(self.wrapper)
        self.assertTrue(request.has_header("Foo"))
        self.assertEqual("bar", request.get_header("Foo"))

        # Override a new field name
        self.wrapper.addCustomHttpHeader("Foo", "bar")
        request = self._get_request(self.wrapper)
        self.assertTrue(request.has_header("Foo"))
        self.assertEqual("bar", request.get_header("Foo"))
        self.wrapper.addCustomHttpHeader("Foo", "bar_2")
        request = self._get_request(self.wrapper)
        self.assertTrue(request.has_header("Foo"))
        self.assertEqual("bar_2", request.get_header("Foo"))

        # Override header field name
        self.wrapper.addCustomHttpHeader("User-agent", "Another UA")
        request = self._get_request(self.wrapper)
        self.assertEqual("Another UA", request.get_header("User-agent"))

    def testClearCustomHttpHeader(self):
        request = self._get_request(self.wrapper)
        self.assertFalse(request.has_header("Foo"))

        # Add new header field name
        self.wrapper.addCustomHttpHeader("Foo_1", "bar_1")
        self.wrapper.addCustomHttpHeader("Foo_2", "bar_2")
        self.wrapper.addCustomHttpHeader("Foo_3", "bar_3")

        self.assertFalse(self.wrapper.clearCustomHttpHeader("Foo_4"))
        self.assertTrue(self.wrapper.clearCustomHttpHeader("Foo_3"))

        customHttpHeaders = self.wrapper.customHttpHeaders

        self.assertTrue("Foo_1" in customHttpHeaders)
        self.assertTrue("Foo_2" in customHttpHeaders)
        self.assertEqual("bar_1", customHttpHeaders["Foo_1"])
        self.assertEqual("bar_2", customHttpHeaders["Foo_2"])

        self.assertFalse(self.wrapper.clearCustomHttpHeader("Foo_3"), "already cleaned")

    def testSetHTTPAuth(self):
        self.assertRaises(TypeError, self.wrapper.setHTTPAuth, 123)
        self.wrapper.setCredentials("login", "password")
        request = self._get_request(self.wrapper)
        self.assertTrue(request.has_header("Authorization"))
        self.assertIsNone(urllib.request._opener)

        self.wrapper.setHTTPAuth(DIGEST)
        self.assertIsNone(urllib.request._opener)
        request = self._get_request(self.wrapper)
        self.assertFalse(request.has_header("Authorization"))
        self.assertEqual(self.wrapper.http_auth, DIGEST)
        self.assertIsInstance(urllib.request._opener, urllib.request.OpenerDirector)

        self.wrapper.setHTTPAuth(DIGEST)
        self.wrapper.setCredentials("login", "password")
        request = self._get_request(self.wrapper)
        self.assertEqual(self.wrapper.http_auth, DIGEST)
        self.assertEqual(self.wrapper.user, "login")
        self.assertEqual(self.wrapper.passwd, "password")
        self.assertEqual(self.wrapper.realm, "SPARQL")
        self.assertNotEqual(self.wrapper.realm, "SPARQL Endpoint")

        self.wrapper.setHTTPAuth(DIGEST)
        self.wrapper.setCredentials("login", "password", realm="SPARQL Endpoint")
        request = self._get_request(self.wrapper)
        self.assertEqual(self.wrapper.http_auth, DIGEST)
        self.assertEqual(self.wrapper.user, "login")
        self.assertEqual(self.wrapper.passwd, "password")
        self.assertEqual(self.wrapper.realm, "SPARQL Endpoint")
        self.assertNotEqual(self.wrapper.realm, "SPARQL")

        self.assertRaises(ValueError, self.wrapper.setHTTPAuth, "OAuth")

        self.wrapper.http_auth = "OAuth"
        self.assertRaises(NotImplementedError, self._get_request, self.wrapper)

    def testSetQuery(self):
        self.wrapper.setQuery(
            "PREFIX example: <http://example.org/INSERT/> SELECT * WHERE {?s ?p ?v}"
        )
        self.assertEqual(SELECT, self.wrapper.queryType)

        self.wrapper.setQuery("PREFIX e: <http://example.org/> INSERT {e:a e:b e:c}")
        self.assertEqual(INSERT, self.wrapper.queryType)

        self.wrapper.setQuery(
            """#CONSTRUCT {?s ?p ?o}
                                   SELECT ?s ?p ?o
                                   WHERE {?s ?p ?o}"""
        )
        self.assertEqual(SELECT, self.wrapper.queryType)

        with warnings.catch_warnings(record=True) as w:
            self.wrapper.setQuery("UNKNOWN {e:a e:b e:c}")
            self.assertEqual(
                SELECT, self.wrapper.queryType, "unknown queries result in SELECT"
            )

    def testSetQueryEncodingIssues(self):
        # further details from issue #35
        query = 'INSERT DATA { <urn:michel> <urn:says> "これはテストです" }'
        query_bytes = query.encode("utf-8")

        self.wrapper.setMethod(POST)
        self.wrapper.setRequestMethod(POSTDIRECTLY)

        self.wrapper.setQuery(query)
        request = self._get_request(self.wrapper)
        self.assertEqual(query_bytes, request.data)

        self.wrapper.setQuery(query_bytes)
        request = self._get_request(self.wrapper)
        self.assertEqual(query_bytes, request.data)

        self.wrapper.setRequestMethod(URLENCODED)

        self.wrapper.setQuery(query)
        parameters = self._get_request_parameters_as_bytes(self.wrapper)
        self.assertEqual(query_bytes, parameters["update"][0])

        self.wrapper.setQuery(query_bytes)
        parameters = self._get_request_parameters_as_bytes(self.wrapper)
        self.assertEqual(query_bytes, parameters["update"][0])

        try:
            self.wrapper.setQuery(query.encode("sjis"))
            self.fail()
        except UnicodeDecodeError:
            self.assertTrue(True)

        try:
            self.wrapper.setQuery({"foo": "bar"})
            self.fail()
        except TypeError:
            self.assertTrue(True)

    def testSetTimeout(self):
        self.wrapper.setTimeout(10)
        self.assertEqual(10, self.wrapper.timeout)

        self.wrapper.resetQuery()
        self.assertEqual(None, self.wrapper.timeout)

    def testClearParameter(self):
        self.wrapper.addParameter("param1", "value1")
        self.wrapper.addParameter("param1", "value2")
        self.wrapper.addParameter("param2", "value2")

        self.assertFalse(self.wrapper.clearParameter("query"))
        self.assertTrue(self.wrapper.clearParameter("param1"))

        pieces = self._get_request_parameters(self.wrapper)

        self.assertFalse("param1" in pieces)
        self.assertTrue("param2" in pieces)
        self.assertEqual(["value2"], pieces["param2"])

        self.assertFalse(self.wrapper.clearParameter("param1"), "already cleaned")

    def testSetMethod(self):
        self.wrapper.setMethod(POST)
        request = self._get_request(self.wrapper)

        self.assertEqual("POST", request.get_method())

        self.wrapper.setMethod(GET)
        request = self._get_request(self.wrapper)

        self.assertEqual("GET", request.get_method())

    def testSetRequestMethod(self):
        self.assertEqual(URLENCODED, self.wrapper.requestMethod)

        self.wrapper.setRequestMethod(POSTDIRECTLY)
        self.assertEqual(POSTDIRECTLY, self.wrapper.requestMethod)

    def testIsSparqlUpdateRequest(self):
        self.wrapper.setQuery("DELETE WHERE {?s ?p ?o}")
        self.assertTrue(self.wrapper.isSparqlUpdateRequest())

        self.wrapper.setQuery("DELETE DATA { <urn:john> <urn:likes> <urn:surfing> }")
        self.assertTrue(self.wrapper.isSparqlUpdateRequest())

        self.wrapper.setQuery(
            """
        PREFIX example: <http://example.org/SELECT/>
        BASE <http://example.org/SELECT>
        DELETE WHERE {?s ?p ?o}
        """
        )
        self.assertTrue(self.wrapper.isSparqlUpdateRequest())

        self.wrapper.setQuery(
            "WITH <urn:graph> DELETE DATA { <urn:john> <urn:likes> <urn:surfing> }"
        )
        self.assertTrue(self.wrapper.isSparqlUpdateRequest())

        self.wrapper.setQuery("INSERT DATA { <urn:john> <urn:likes> <urn:surfing> }")
        self.assertTrue(self.wrapper.isSparqlUpdateRequest())

        self.wrapper.setQuery(
            "WITH <urn:graph> INSERT DATA { <urn:john> <urn:likes> <urn:surfing> }"
        )
        self.assertTrue(self.wrapper.isSparqlUpdateRequest())

        self.wrapper.setQuery("CREATE GRAPH <urn:graph>")
        self.assertTrue(self.wrapper.isSparqlUpdateRequest())

        self.wrapper.setQuery("CLEAR GRAPH <urn:graph>")
        self.assertTrue(self.wrapper.isSparqlUpdateRequest())

        self.wrapper.setQuery("DROP GRAPH <urn:graph>")
        self.assertTrue(self.wrapper.isSparqlUpdateRequest())

        self.wrapper.setQuery("MOVE GRAPH <urn:graph1> TO GRAPH <urn:graph2>")
        self.assertTrue(self.wrapper.isSparqlUpdateRequest())

        self.wrapper.setQuery("LOAD <http://localhost/file.rdf> INTO GRAPH <urn:graph>")
        self.assertTrue(self.wrapper.isSparqlUpdateRequest())

        self.wrapper.setQuery("COPY <urn:graph1> TO GRAPH <urn:graph2>")
        self.assertTrue(self.wrapper.isSparqlUpdateRequest())

        self.wrapper.setQuery("ADD <urn:graph1> TO GRAPH <urn:graph2>")
        self.assertTrue(self.wrapper.isSparqlUpdateRequest())

    def testIsSparqlQueryRequest(self):
        self.wrapper.setQuery("SELECT * WHERE {?s ?p ?o}")
        self.assertTrue(self.wrapper.isSparqlQueryRequest())

        self.wrapper.setQuery(
            """
        PREFIX example: <http://example.org/DELETE/>
        BASE <http://example.org/MODIFY>
        ASK WHERE {?s ?p ?o}
        """
        )
        self.assertTrue(self.wrapper.isSparqlQueryRequest())
        self.assertFalse(self.wrapper.isSparqlUpdateRequest())

    def testQuery(self):
        qr = self.wrapper.query()
        self.assertTrue(isinstance(qr, QueryResult))

        request = qr.response.request  # possible due to mock above
        self.assertTrue(isinstance(request, Request))

        parameters = self._get_parameters_from_request(request)
        self.assertTrue("query" in parameters)
        self.assertTrue("update" not in parameters)

        self.wrapper.setMethod(POST)
        self.wrapper.setQuery("PREFIX e: <http://example.org/> INSERT {e:a e:b e:c}")
        parameters = self._get_request_parameters(self.wrapper)
        self.assertTrue("update" in parameters)
        self.assertTrue("query" not in parameters)
        # _returnFormatSetting = ["format", "output", "results"]
        self.assertTrue("format" not in parameters)
        self.assertTrue("output" not in parameters)
        self.assertTrue("results" not in parameters)

        _victim.urlopener = urlopener_error_generator(400)
        try:
            self.wrapper.query()
            self.fail("should have raised exception")
        except QueryBadFormed as e:
            #  TODO: check exception-format
            pass
        except:
            self.fail("got wrong exception")

        _victim.urlopener = urlopener_error_generator(401)
        try:
            self.wrapper.query()
            self.fail("should have raised exception")
        except Unauthorized as e:
            #  TODO: check exception-format
            pass
        except:
            self.fail("got wrong exception")

        _victim.urlopener = urlopener_error_generator(404)
        try:
            self.wrapper.query()
            self.fail("should have raised exception")
        except EndPointNotFound as e:
            #  TODO: check exception-format
            pass
        except:
            self.fail("got wrong exception")

        _victim.urlopener = urlopener_error_generator(414)
        try:
            self.wrapper.query()
            self.fail("should have raised exception")
        except URITooLong as e:
            #  TODO: check exception-format
            pass
        except:
            self.fail("got wrong exception")

        _victim.urlopener = urlopener_error_generator(500)
        try:
            self.wrapper.query()
            self.fail("should have raised exception")
        except EndPointInternalError as e:
            #  TODO: check exception-format
            pass
        except:
            self.fail("got wrong exception")

        _victim.urlopener = urlopener_error_generator(999)
        try:
            self.wrapper.query()
            self.fail("should have raised exception")
        except HTTPError as e:
            #  TODO: check exception-format
            pass
        except:
            self.fail("got wrong exception")

    def testQueryEncoding(self):
        query = 'INSERT DATA { <urn:michel> <urn:says> "é" }'

        wrapper = SPARQLWrapper("http://example.com:3030/example")
        wrapper.setMethod(POST)
        wrapper.setRequestMethod(URLENCODED)
        wrapper.setQuery(query)

        _victim.urlopener = urlopener_check_data_encoding
        wrapper.query()

    def testQueryAndConvert(self):
        _oldQueryResult = _victim.QueryResult

        class FakeQueryResult(object):
            def __init__(self, result):
                pass

            def convert(self):
                return True

        try:
            _victim.QueryResult = FakeQueryResult
            result = self.wrapper.queryAndConvert()
            self.assertEqual(True, result)
        finally:
            _victim.QueryResult = _oldQueryResult

    def testComments(self):
        # see issue #32
        self.wrapper.setQuery(
            """
# this is a comment
select * where { ?s ?p ?o }
"""
        )
        self.assertTrue(self.wrapper.isSparqlQueryRequest())

    def testHashInPrefixes(self):
        # see issue #77
        self.wrapper.setQuery(
            """
PREFIX rdfs:    <http://www.w3.org/2000/01/rdf-schema#>
select * where { ?s ?p ?o }
"""
        )
        self.assertTrue(self.wrapper.isSparqlQueryRequest())

    def testHashInPrefixComplex(self):
        # see issue #77
        self.wrapper.setQuery(
            """
PREFIX rdf:     <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX rdfs:    <http://www.w3.org/2000/01/rdf-schema#>
PREFIX weather: <http://hal.zamia.org/weather/>
PREFIX dbo:     <http://dbpedia.org/ontology/> 
PREFIX dbr:     <http://dbpedia.org/resource/> 
PREFIX dbp:     <http://dbpedia.org/property/> 
PREFIX xml:     <http://www.w3.org/XML/1998/namespace> 
PREFIX xsd:     <http://www.w3.org/2001/XMLSchema#> 

SELECT DISTINCT ?location ?cityid ?timezone ?label
WHERE {
  ?location weather:cityid ?cityid .
  ?location weather:timezone ?timezone .
  ?location rdfs:label ?label .
}
"""
        )
        self.assertTrue(self.wrapper.isSparqlQueryRequest())

    def testHashWithNoComments(self):
        # see issue #77
        query = """
PREFIX rdf:     <http://www.w3.org/1999/02/22-rdf-syntax-ns#>

SELECT *
WHERE {
  ?s ?p ?o .
}
"""
        parsed_query = self.wrapper._cleanComments(query)
        self.assertEqual(query, parsed_query)

    def testCommentBeginningLine(self):
        # see issue #77
        query = """
PREFIX rdf:     <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
# a comment
SELECT *
WHERE {
  ?s ?p ?o .
}
"""
        expected_parsed_query = """
PREFIX rdf:     <http://www.w3.org/1999/02/22-rdf-syntax-ns#>

SELECT *
WHERE {
  ?s ?p ?o .
}
"""
        parsed_query = self.wrapper._cleanComments(query)
        self.assertEqual(expected_parsed_query, parsed_query)

    def testCommentEmtpyLine(self):
        # see issue #77
        query = """
PREFIX rdf:     <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
     # a comment
SELECT *
WHERE {
  ?s ?p ?o .
}
"""
        expected_parsed_query = """
PREFIX rdf:     <http://www.w3.org/1999/02/22-rdf-syntax-ns#>

SELECT *
WHERE {
  ?s ?p ?o .
}
"""
        parsed_query = self.wrapper._cleanComments(query)
        self.assertEqual(expected_parsed_query, parsed_query)

    def testCommentsFirstLine(self):
        # see issue #77
        query = """#CONSTRUCT {?s ?p ?o}
                                   SELECT ?s ?p ?o
                                   WHERE {?s ?p ?o}"""
        expected_parsed_query = """

                                   SELECT ?s ?p ?o
                                   WHERE {?s ?p ?o}"""

        parsed_query = self.wrapper._cleanComments(query)
        self.assertEqual(expected_parsed_query, parsed_query)

    @unittest.skip("issue #80")
    def testCommentAfterStatements(self):
        # see issue #77
        query = """
PREFIX rdf:     <http://www.w3.org/1999/02/22-rdf-syntax-ns#>

SELECT *
WHERE {     # this is the where condition
  ?s ?p ?o .
}
"""
        expected_parsed_query = """
PREFIX rdf:     <http://www.w3.org/1999/02/22-rdf-syntax-ns#>

SELECT *
WHERE {
  ?s ?p ?o .
}
"""
        parsed_query = self.wrapper._cleanComments(query)
        self.assertEqual(expected_parsed_query, parsed_query)

    def testSingleLineQueryLine(self):
        # see issue #74
        query = "prefix whatever: <http://example.org/blah#> ASK { ?s ?p ?o }"
        parsed_query = self.wrapper._cleanComments(query)
        self.assertEqual(query, parsed_query)

        self.wrapper.setQuery(query)
        self.assertTrue(self.wrapper.isSparqlQueryRequest())

    def testOnlyConneg(self):
        # see issue #82
        query = "prefix whatever: <http://example.org/blah#> ASK { ?s ?p ?o }"
        self.wrapper.setOnlyConneg(False)
        self.wrapper.setQuery(query)
        request = self._get_request(self.wrapper)
        request_params = dict(parse_qsl(urlparse(request.get_full_url()).query))
        for returnFormatSetting in [
            "format",
            "output",
            "results",
        ]:  # Obviously _returnFormatSetting is not accessible from SPARQLWrapper, so we copy&paste the possible values
            self.assertTrue(
                returnFormatSetting in request_params,
                "URL parameter '%s' was not sent, and it was expected"
                % returnFormatSetting,
            )

        # ONLY Content Negotiation
        self.wrapper.resetQuery()
        self.wrapper.setOnlyConneg(True)
        self.wrapper.setQuery(query)
        request = self._get_request(self.wrapper)
        request_params = dict(parse_qsl(urlparse(request.get_full_url()).query))
        for returnFormatSetting in [
            "format",
            "output",
            "results",
        ]:  # Obviously _returnFormatSetting is not accessible from SPARQLWrapper, so we copy&paste the possible values
            self.assertFalse(
                returnFormatSetting in request_params,
                "URL parameter '%s' was sent, and it was not expected (only Content Negotiation)"
                % returnFormatSetting,
            )


class QueryResult_Test(unittest.TestCase):
    def testConstructor(self):
        qr = QueryResult("result")
        self.assertEqual("result", qr.response)
        try:
            format = qr.requestedFormat
            self.fail("format is not supposed to be set")
        except:
            pass

        qr = QueryResult(("result", "format"))
        self.assertEqual("result", qr.response)
        self.assertEqual("format", qr.requestedFormat)

    def testProxyingToResponse(self):
        class FakeResponse(object):
            def __init__(self):
                self.geturl_called = False
                self.info_called = False
                self.iter_called = False
                self.next_called = False

            def geturl(self):
                self.geturl_called = True

            def info(self):
                self.info_called = True
                return {"key": "value"}

            def __iter__(self):
                self.iter_called = True

            def __next__(self):
                self.next_called = True

        result = FakeResponse()

        qr = QueryResult(result)
        qr.geturl()
        qr.__iter__()
        next(qr)

        self.assertTrue(result.geturl_called)
        self.assertTrue(result.iter_called)
        self.assertTrue(result.next_called)

        info = qr.info()
        self.assertTrue(result.info_called)
        self.assertEqual(
            "value", info.__getitem__("KEY"), "keys should be case-insensitive"
        )

    def testConvert(self):
        class FakeResponse(object):
            def __init__(self, content_type):
                self.content_type = content_type

            def info(self):
                return {"content-type": self.content_type}

            def read(self, len):
                return ""

        def _mime_vs_type(mime, requested_type):
            """
            :param mime: mimetype/Content-Type of the response
            :param requested_type: requested mimetype (alias)
            :return: number of warnings produced by combo
            """
            with warnings.catch_warnings(record=True) as w:
                qr = QueryResult((FakeResponse(mime), requested_type))

                try:
                    qr.convert()
                except:
                    pass

                # if len(w) > 0: print(w[0].message) # FOR DEBUG
                # if len(w) > 1: print(w[1].message) # FOR DEBUG

                return len(w)

        # In the cases of "application/ld+json" and "application/rdf+xml", the
        # RDFLib raised a warning because the manually created QueryResult has no real
        # response value (implemented a fake read).
        # "WARNING:rdflib.term:  does not look like a valid URI, trying to serialize this will break."
        self.assertEqual(0, _mime_vs_type("application/sparql-results+xml", XML))
        self.assertEqual(0, _mime_vs_type("application/sparql-results+json", JSON))
        self.assertEqual(0, _mime_vs_type("text/n3", N3))
        self.assertEqual(0, _mime_vs_type("text/turtle", TURTLE))
        self.assertEqual(0, _mime_vs_type("application/turtle", TURTLE))
        self.assertEqual(0, _mime_vs_type("application/json", JSON))

        self.assertEqual(0, _mime_vs_type("application/ld+json", JSONLD))
        self.assertEqual(0, _mime_vs_type("application/rdf+xml", XML))
        self.assertEqual(0, _mime_vs_type("application/rdf+xml", RDF))
        self.assertEqual(0, _mime_vs_type("application/rdf+xml", RDFXML))

        self.assertEqual(0, _mime_vs_type("text/csv", CSV))
        self.assertEqual(0, _mime_vs_type("text/tab-separated-values", TSV))
        self.assertEqual(0, _mime_vs_type("application/xml", XML))

        # unknown response content type 'application/x-foo-bar' returning raw response...
        self.assertEqual(1, _mime_vs_type("application/x-foo-bar", XML), "invalid mime")

        # Format requested was xxx, but yyy (zzz) has been returned by the endpoint
        self.assertEqual(1, _mime_vs_type("application/sparql-results+xml", N3))
        self.assertEqual(1, _mime_vs_type("application/sparql-results+json", XML))
        self.assertEqual(1, _mime_vs_type("text/n3", JSON))
        self.assertEqual(1, _mime_vs_type("text/turtle", XML))

        # Format requested was xxx, but yyy (zzz) has been returned by the endpoint
        self.assertEqual(1, _mime_vs_type("application/ld+json", XML))  # Warning
        self.assertEqual(1, _mime_vs_type("application/ld+json", N3))  # Warning
        self.assertEqual(1, _mime_vs_type("application/rdf+xml", JSON))  # Warning
        self.assertEqual(1, _mime_vs_type("application/rdf+xml", N3))  # Warning

    def testPrint_results(self):
        """
        print_results() is only allowed for JSON return format.
        """

        class FakeResponse(object):
            def __init__(self, content_type):
                self.content_type = content_type

            def info(self):
                return {"content-type": self.content_type}

            def read(self, len):
                return ""

        def _print_results(mime):
            """
            :param mime: mimetype/Content-Type of the response
            :return: number of warnings produced by combo
            """
            with warnings.catch_warnings(record=True) as w:
                qr = QueryResult(FakeResponse(mime))

                try:
                    qr.print_results()
                except:
                    pass

                # if len(w) > 1: print(w[1].message) # FOR DEBUG

                return len(w)

        self.assertEqual(0, _print_results("application/sparql-results+json"))
        self.assertEqual(0, _print_results("application/json"))
        self.assertEqual(0, _print_results("text/javascript"))
        self.assertEqual(0, _print_results("application/javascript"))

        # Format return was xxx, but yyy was expected. No printing.
        self.assertEqual(1, _print_results("application/sparql-results+xml"))
        self.assertEqual(1, _print_results("application/xml"))
        self.assertEqual(1, _print_results("application/rdf+xml"))

        self.assertEqual(1, _print_results("application/turtle"))
        self.assertEqual(1, _print_results("text/turtle"))

        self.assertEqual(1, _print_results("text/rdf+n3"))
        self.assertEqual(1, _print_results("application/n-triples"))
        self.assertEqual(1, _print_results("application/n3"))
        self.assertEqual(1, _print_results("text/n3"))

        self.assertEqual(1, _print_results("text/csv"))

        self.assertEqual(1, _print_results("text/tab-separated-values"))

        self.assertEqual(1, _print_results("application/ld+json"))
        self.assertEqual(1, _print_results("application/x-json+ld"))

        # Unknown response content type. Returning raw content-type ('application/x-foo-bar').
        # Format return was application/x-foo-bar, but JSON was expected. No printing.
        self.assertEqual(2, _print_results("application/x-foo-bar"))


class QueryType_Time_Test(unittest.TestCase):
    def testQueries(self):
        sparql = SPARQLWrapper("http://example.org/sparql")

        queries = []

        queries.append(
            """
    PREFIX a: <http://dbpedia.org/a>
    PREFIX b: <http://dbpedia.org/b>
    PREFIX c: <http://dbpedia.org/c>
    PREFIX d: <http://dbpedia.org/d>
    PREFIX e: <http://dbpedia.org/e>
    PREFIX f: <http://dbpedia.org/f>
    PREFIX g: <http://dbpedia.org/g>
    PREFIX h: <http://dbpedia.org/h>
    FROM <http://dbpedia.org>
    SELECT ?s ?p ?o WHERE {
        ?s ?p ?o.
    }"""
        )

        queries.append(
            """PREFIX a: <http://dbpedia.org/a>
    PREFIX b: <http://dbpedia.org/b>
    PREFIX c: <http://dbpedia.org/c>
    PREFIX d: <http://dbpedia.org/d>
    PREFIX e: <http://dbpedia.org/e>
    PREFIX f: <http://dbpedia.org/f>
    PREFIX g: <http://dbpedia.org/g>
    PREFIX h: <http://dbpedia.org/h>
    FROM <http://dbpedia.org>
    SELECT ?s ?p ?o WHERE {
        ?s ?p ?o.
    }"""
        )

        queries.append(
            """PREFIX a: <http://dbpedia.org/a>
PREFIX b: <http://dbpedia.org/b>
PREFIX c: <http://dbpedia.org/c>
PREFIX d: <http://dbpedia.org/d>
PREFIX e: <http://dbpedia.org/e>
PREFIX f: <http://dbpedia.org/f>
PREFIX g: <http://dbpedia.org/g>
PREFIX h: <http://dbpedia.org/h>
FROM <http://dbpedia.org>
SELECT ?s ?p ?o WHERE {
    ?s ?p ?o.
}"""
        )

        queries.append(
            """PREFIX a: <http://dbpedia.org/a>
PREFIX b: <http://dbpedia.org/b>
PREFIX c: <http://dbpedia.org/c>
PREFIX d: <http://dbpedia.org/d>
PREFIX e: <http://dbpedia.org/e>
PREFIX f: <http://dbpedia.org/f>
PREFIX g: <http://dbpedia.org/g>
PREFIX h: <http://dbpedia.org/h>
SELECT ?s ?p ?o WHERE {
    ?s ?p ?o.
}"""
        )

        queries.append(
            """
    PREFIX a: <http://dbpedia.org/a>
    PREFIX b: <http://dbpedia.org/b>
    PREFIX c: <http://dbpedia.org/c>
    PREFIX d: <http://dbpedia.org/d>
    PREFIX e: <http://dbpedia.org/e>
    PREFIX f: <http://dbpedia.org/f>
    PREFIX g: <http://dbpedia.org/g>
    PREFIX h: <http://dbpedia.org/h>
    SELECT ?s ?p ?o WHERE {
        ?s ?p ?o.
    }"""
        )

        queries.append(
            """
    FROM <http://dbpedia.org>
    SELECT ?s ?p ?o WHERE {
        ?s ?p ?o.
    }"""
        )

        queries.append(
            """
    PREFIX a: <http://dbpedia.org/a>
    PREFIX b: <http://dbpedia.org/b>
    PREFIX c: <http://dbpedia.org/c>
    PREFIX d: <http://dbpedia.org/d>
    PREFIX e: <http://dbpedia.org/e>
    PREFIX f: <http://dbpedia.org/f>
    PREFIX g: <http://dbpedia.org/g>
    PREFIX h: <http://dbpedia.org/h>
    PREFIX a: <http://dbpedia.org/a>
    PREFIX b: <http://dbpedia.org/b>
    PREFIX c: <http://dbpedia.org/c>
    PREFIX d: <http://dbpedia.org/d>
    PREFIX e: <http://dbpedia.org/e>
    PREFIX f: <http://dbpedia.org/f>
    PREFIX g: <http://dbpedia.org/g>
    PREFIX h: <http://dbpedia.org/h>
    PREFIX a: <http://dbpedia.org/a>
    PREFIX b: <http://dbpedia.org/b>
    PREFIX c: <http://dbpedia.org/c>
    PREFIX d: <http://dbpedia.org/d>
    PREFIX e: <http://dbpedia.org/e>
    PREFIX f: <http://dbpedia.org/f>
    PREFIX g: <http://dbpedia.org/g>
    PREFIX h: <http://dbpedia.org/h>
    FROM <http://dbpedia.org>
    SELECT ?s ?p ?o WHERE {
        ?s ?p ?o.
    }"""
        )

        queries.append(
            """PREFIX a: <http://dbpedia.org/a>
    PREFIX b: <http://dbpedia.org/b>
    PREFIX c: <http://dbpedia.org/c>
    PREFIX d: <http://dbpedia.org/d>
    PREFIX e: <http://dbpedia.org/e>
    PREFIX f: <http://dbpedia.org/f>
    PREFIX g: <http://dbpedia.org/g>
    PREFIX h: <http://dbpedia.org/h>
    PREFIX a: <http://dbpedia.org/a>
    PREFIX b: <http://dbpedia.org/b>
    PREFIX c: <http://dbpedia.org/c>
    PREFIX d: <http://dbpedia.org/d>
    PREFIX e: <http://dbpedia.org/e>
    PREFIX f: <http://dbpedia.org/f>
    PREFIX g: <http://dbpedia.org/g>
    PREFIX h: <http://dbpedia.org/h>
    PREFIX a: <http://dbpedia.org/a>
    PREFIX b: <http://dbpedia.org/b>
    PREFIX c: <http://dbpedia.org/c>
    PREFIX d: <http://dbpedia.org/d>
    PREFIX e: <http://dbpedia.org/e>
    PREFIX f: <http://dbpedia.org/f>
    PREFIX g: <http://dbpedia.org/g>
    PREFIX h: <http://dbpedia.org/h>
    FROM <http://dbpedia.org>
    SELECT ?s ?p ?o WHERE {
        ?s ?p ?o.
    }"""
        )

        for query in queries:
            start_time = time.time()
            sparql.setQuery(query)
            self.assertTrue(
                (time.time() - start_time) < 0.001
            )  # less than 0.001 second


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

import inspect
import io
import os
import sys
import textwrap
import unittest

# prefer local copy to the one which is installed
# hack from http://stackoverflow.com/a/6098238/280539
_top_level_path = os.path.realpath(
    os.path.abspath(
        os.path.join(os.path.split(inspect.getfile(inspect.currentframe()))[0], "..")
    )
)
if _top_level_path not in sys.path:
    sys.path.insert(0, _top_level_path)
# end of hack

from SPARQLWrapper.main import main, parse_args
from SPARQLWrapper import POST

endpoint = "http://ja.dbpedia.org/sparql"
testfile = os.path.join(os.path.dirname(__file__), "test.rq")
testquery = "SELECT DISTINCT ?x WHERE { ?x ?y ?z . } LIMIT 1"


class SPARQLWrapperCLI_Test_Base(unittest.TestCase):
    def setUp(self):
        self.org_stdout, sys.stdout = sys.stdout, io.StringIO()
        self.org_stderr, sys.stderr = sys.stderr, io.StringIO()

    def tearDown(self):
        sys.stdout = self.org_stdout


class SPARQLWrapperCLIParser_Test(SPARQLWrapperCLI_Test_Base):
    def testHelp(self):
        with self.assertRaises(SystemExit) as cm:
            parse_args(["-h"])

        self.assertEqual(cm.exception.code, 0)
        self.assertEqual(sys.stdout.getvalue()[:5], "usage")

    def testVersion(self):
        with self.assertRaises(SystemExit) as cm:
            parse_args(["-V"])

        self.assertEqual(cm.exception.code, 0)
        self.assertEqual(sys.stdout.getvalue()[:3], "rqw")

    def testNoarg(self):
        with self.assertRaises(SystemExit) as cm:
            parse_args([])

        self.assertEqual(cm.exception.code, 2)
        self.assertEqual(
            sys.stderr.getvalue().split("\n")[1],
            "rqw: error: one of the arguments -f/--file -Q/--query is required",
        )

    def testQueryAndFile(self):
        with self.assertRaises(SystemExit) as cm:
            parse_args(["-Q", testquery, "-f", "-"])

        self.assertEqual(cm.exception.code, 2)
        self.assertEqual(
            sys.stderr.getvalue().split("\n")[1],
            "rqw: error: argument -f/--file: not allowed with argument -Q/--query",
        )

    def testInvalidFormat(self):
        with self.assertRaises(SystemExit) as cm:
            parse_args(["-Q", testquery, "-F", "jjssoonn"])

        self.assertEqual(cm.exception.code, 2)
        self.assertEqual(
            sys.stderr.getvalue().split("\n")[1],
            "rqw: error: argument -F/--format: invalid choice: 'jjssoonn' (choose from 'json', 'xml', 'turtle', 'n3', 'rdf', 'rdf+xml', 'csv', 'tsv', 'json-ld')",
        )

    def testInvalidFile(self):
        with self.assertRaises(SystemExit) as cm:
            parse_args(["-f", "440044.rq"])

        self.assertEqual(cm.exception.code, 2)
        self.assertEqual(
            sys.stderr.getvalue().split("\n")[1],
            "rqw: error: argument -f/--file: file '440044.rq' is not found",
        )


class SPARQLWrapperCLI_Test(SPARQLWrapperCLI_Test_Base):
    def testQueryWithEndpoint(self):
        main(
            [
                "-Q",
                testquery,
                "-e",
                endpoint,
            ]
        )

        self.assertEqual(
            sys.stdout.getvalue(),
            textwrap.dedent(
                """\
            {
                "head": {
                    "link": [],
                    "vars": [
                        "x"
                    ]
                },
                "results": {
                    "distinct": false,
                    "ordered": true,
                    "bindings": [
                        {
                            "x": {
                                "type": "uri",
                                "value": "http://www.openlinksw.com/virtrdf-data-formats#default-iid"
                            }
                        }
                    ]
                }
            }
            """
            ),
        )

    def testQueryWithFile(self):
        main(["-f", testfile, "-e", endpoint])

        self.assertEqual(
            sys.stdout.getvalue(),
            textwrap.dedent(
                """\
            {
                "head": {
                    "link": [],
                    "vars": [
                        "pllabel"
                    ]
                },
                "results": {
                    "distinct": false,
                    "ordered": true,
                    "bindings": [
                        {
                            "pllabel": {
                                "type": "literal",
                                "xml:lang": "ja",
                                "value": "PARLOG"
                            }
                        }
                    ]
                }
            }
            """
            ),
        )

    def testQueryWithFileXML(self):
        main(["-f", testfile, "-e", endpoint, "-F", "xml"])

        self.assertEqual(
            sys.stdout.getvalue(),
            textwrap.dedent(
                """\
            <?xml version="1.0" ?><sparql xmlns="http://www.w3.org/2005/sparql-results#" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://www.w3.org/2001/sw/DataAccess/rf1/result2.xsd">
             <head>
              <variable name="pllabel"/>
             </head>
             <results distinct="false" ordered="true">
              <result>
               <binding name="pllabel"><literal xml:lang="ja">PARLOG</literal></binding>
              </result>
             </results>
            </sparql>
            """
            ),
        )

    def testQueryWithFileTurtle(self):
        main(["-f", testfile, "-e", endpoint, "-F", "turtle"])

        self.assertEqual(
            sys.stdout.getvalue(),
            textwrap.dedent(
                """\
            @prefix res: <http://www.w3.org/2005/sparql-results#> .
            @prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
            _:_ a res:ResultSet .
            _:_ res:resultVariable "pllabel" .
            _:_ res:solution [
                  res:binding [ res:variable "pllabel" ; res:value "PARLOG"@ja ] ] .\n
            """
            ),
        )

    def testQueryWithFileTurtleQuiet(self):
        main(
            [
                "-f",
                testfile,
                "-e",
                endpoint,
                "-F",
                "turtle",
                "-q",
            ]
        )
        self.assertEqual(sys.stderr.getvalue(), "")
        self.assertEqual(
            sys.stdout.getvalue(),
            textwrap.dedent(
                """\
            @prefix res: <http://www.w3.org/2005/sparql-results#> .
            @prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
            _:_ a res:ResultSet .
            _:_ res:resultVariable "pllabel" .
            _:_ res:solution [
                  res:binding [ res:variable "pllabel" ; res:value "PARLOG"@ja ] ] .\n
            """
            ),
        )

    def testQueryWithFileN3(self):
        main(["-f", testfile, "-e", endpoint, "-F", "n3"])

        self.assertEqual(
            sys.stdout.getvalue(),
            textwrap.dedent(
                """\
            @prefix res: <http://www.w3.org/2005/sparql-results#> .
            @prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
            _:_ a res:ResultSet .
            _:_ res:resultVariable "pllabel" .
            _:_ res:solution [
                  res:binding [ res:variable "pllabel" ; res:value "PARLOG"@ja ] ] .\n
            """
            ),
        )

    def testQueryRDF(self):
        main(["-Q", "DESCRIBE <http://ja.wikipedia.org/wiki/SPARQL>", "-e", endpoint, "-F", "rdf"])

        self.assertEqual(
            sys.stdout.getvalue(),
            textwrap.dedent(
                """\
            @prefix dc: <http://purl.org/dc/elements/1.1/> .
            @prefix foaf: <http://xmlns.com/foaf/0.1/> .

            <http://ja.dbpedia.org/resource/SPARQL> foaf:isPrimaryTopicOf <http://ja.wikipedia.org/wiki/SPARQL> .

            <http://ja.wikipedia.org/wiki/SPARQL> a foaf:Document ;
                dc:language "ja" ;
                foaf:primaryTopic <http://ja.dbpedia.org/resource/SPARQL> .


            """
            ),
        )

    def testQueryWithFileRDFXML(self):
        main(["-f", testfile, "-e", endpoint, "-F", "rdf+xml"])

        self.assertEqual(
            sys.stdout.getvalue(),
            textwrap.dedent(
                """\
            <?xml version="1.0" ?><sparql xmlns="http://www.w3.org/2005/sparql-results#" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://www.w3.org/2001/sw/DataAccess/rf1/result2.xsd">
             <head>
              <variable name="pllabel"/>
             </head>
             <results distinct="false" ordered="true">
              <result>
               <binding name="pllabel"><literal xml:lang="ja">PARLOG</literal></binding>
              </result>
             </results>
            </sparql>
            """
            ),
        )

    def testQueryWithFileCSV(self):
        main(["-f", testfile, "-e", endpoint, "-F", "csv"])

        self.assertEqual(
            sys.stdout.getvalue(),
            textwrap.dedent(
                """\
                "pllabel"
                "PARLOG"\n
            """
            ),
        )

    def testQueryWithFileTSV(self):
        main(["-f", testfile, "-e", endpoint, "-F", "tsv"])

        self.assertEqual(
            sys.stdout.getvalue(),
            textwrap.dedent(
                """\
                "pllabel"
                "PARLOG"\n
            """
            ),
        )

    def testQueryToLovFuseki(self):
        main(["-e", "https://lov.linkeddata.es/dataset/lov/sparql/", "-Q", testquery])
        self.assertEqual(
            sys.stdout.getvalue(),
            textwrap.dedent(
                """\
            {
                "head": {
                    "vars": [
                        "x"
                    ]
                },
                "results": {
                    "bindings": [
                        {
                            "x": {
                                "type": "uri",
                                "value": "http://www.w3.org/2002/07/owl#someValuesFrom"
                            }
                        }
                    ]
                }
            }
            """
            ),
        )

    def testQueryToRDF4J(self):
        main(
            [
                "-e",
                "http://vocabs.ands.org.au/repository/api/sparql/csiro_international-chronostratigraphic-chart_2018-revised-corrected",
                "-Q",
                testquery,
            ]
        )
        self.assertEqual(
            sys.stdout.getvalue(),
            textwrap.dedent(
                """\
            {
                "head": {
                    "vars": [
                        "x"
                    ]
                },
                "results": {
                    "bindings": [
                        {
                            "x": {
                                "type": "uri",
                                "value": "http://www.w3.org/1999/02/22-rdf-syntax-ns#type"
                            }
                        }
                    ]
                }
            }
            """
            ),
        )

    def testQueryToAllegroGraph(self):
        main(["-e", "https://mmisw.org/sparql", "-Q", testquery])
        self.assertEqual(
            sys.stdout.getvalue(),
            textwrap.dedent(
                """\
            {
                "head": {
                    "vars": [
                        "x"
                    ]
                },
                "results": {
                    "bindings": [
                        {
                            "x": {
                                "type": "uri",
                                "value": "https://mmisw.org/ont/~mjuckes/cmip_variables_alpha/rsdcs4co2"
                            }
                        }
                    ]
                }
            }
            """
            ),
        )

    def testQueryToGraphDBEnterprise(self):
        main(["-e", "http://factforge.net/repositories/ff-news", "-Q", testquery])
        self.assertEqual(
            sys.stdout.getvalue(),
            textwrap.dedent(
                """\
            {
                "head": {
                    "vars": [
                        "x"
                    ]
                },
                "results": {
                    "bindings": [
                        {
                            "x": {
                                "type": "uri",
                                "value": "http://www.w3.org/1999/02/22-rdf-syntax-ns#type"
                            }
                        }
                    ]
                }
            }
            """
            ),
        )

    def testQueryToStardog(self):
        main(["-e", "https://lindas.admin.ch/query", "-Q", testquery, "-m", POST])
        self.assertEqual(
            sys.stdout.getvalue(),
            textwrap.dedent(
                """\
            {
                "head": {
                    "vars": [
                        "x"
                    ]
                },
                "results": {
                    "bindings": [
                        {
                            "x": {
                                "type": "uri",
                                "value": "http://classifications.data.admin.ch/canton/bl"
                            }
                        }
                    ]
                }
            }
            """
            ),
        )

    def testQueryToAgrovoc_AllegroGraph(self):
        main(["-e", "https://agrovoc.fao.org/sparql", "-Q", testquery])
        self.assertEqual(
            sys.stdout.getvalue(),
            textwrap.dedent(
                """\
            {
                "head": {
                    "vars": [
                        "x"
                    ]
                },
                "results": {
                    "bindings": [
                        {
                            "x": {
                                "type": "uri",
                                "value": "http://aims.fao.org/aos/agrovoc/"
                            }
                        }
                    ]
                }
            }
            """
            ),
        )

    def testQueryToVirtuosoV8(self):
        main(["-e", "http://dbpedia-live.openlinksw.com/sparql", "-Q", testquery])
        self.assertEqual(
            sys.stdout.getvalue(),
            textwrap.dedent(
                """\
            {
                "head": {
                    "link": [],
                    "vars": [
                        "x"
                    ]
                },
                "results": {
                    "distinct": false,
                    "ordered": true,
                    "bindings": [
                        {
                            "x": {
                                "type": "uri",
                                "value": "http://www.openlinksw.com/virtrdf-data-formats#default-iid"
                            }
                        }
                    ]
                }
            }
            """
            ),
        )

    def testQueryToVirtuosoV7(self):
        main(["-e", "http://dbpedia.org/sparql", "-Q", testquery])
        self.assertEqual(
            sys.stdout.getvalue(),
            textwrap.dedent(
                """\
            {
                "head": {
                    "link": [],
                    "vars": [
                        "x"
                    ]
                },
                "results": {
                    "distinct": false,
                    "ordered": true,
                    "bindings": [
                        {
                            "x": {
                                "type": "uri",
                                "value": "http://www.openlinksw.com/virtrdf-data-formats#default-iid"
                            }
                        }
                    ]
                }
            }
            """
            ),
        )

    def testQueryToBrazeGraph(self):
        main(["-e", "https://query.wikidata.org/sparql", "-Q", testquery])
        self.assertEqual(
            sys.stdout.getvalue(),
            textwrap.dedent(
                """\
            {
                "head": {
                    "vars": [
                        "x"
                    ]
                },
                "results": {
                    "bindings": [
                        {
                            "x": {
                                "type": "uri",
                                "value": "http://wikiba.se/ontology#Dump"
                            }
                        }
                    ]
                }
            }
            """
            ),
        )

    def testQueryToFuseki2V3_6(self):
        main(["-e", "https://agrovoc.uniroma2.it/sparql/", "-Q", testquery])
        self.assertEqual(
            sys.stdout.getvalue(),
            textwrap.dedent(
                """\
            {
                "head": {
                    "vars": [
                        "x"
                    ]
                },
                "results": {
                    "bindings": [
                        {
                            "x": {
                                "type": "uri",
                                "value": "http://aims.fao.org/aos/agrovoc/"
                            }
                        }
                    ]
                }
            }
            """
            ),
        )

    def testQueryToFuseki2V3_8(self):
        main(["-e", "http://zbw.eu/beta/sparql/stw/query", "-Q", testquery])
        self.assertEqual(
            sys.stdout.getvalue(),
            textwrap.dedent(
                """\
            {
                "head": {
                    "vars": [
                        "x"
                    ]
                },
                "results": {
                    "bindings": [
                        {
                            "x": {
                                "type": "uri",
                                "value": "http://www.w3.org/2004/02/skos/core"
                            }
                        }
                    ]
                }
            }
            """
            ),
        )

    def testQueryTo4store(self):
        main(["-e", "http://rdf.chise.org/sparql", "-Q", testquery])
        self.assertEqual(
            sys.stdout.getvalue(),
            textwrap.dedent(
                """\
            {
                "head": {
                    "vars": [
                        "x"
                    ]
                },
                "results": {
                    "bindings": [
                        {
                            "x": {
                                "type": "bnode",
                                "value": "b1f4d352f000000fc"
                            }
                        }
                    ]
                }
            }
            """
            ),
        )



# -*- coding: utf8 -*-

"""

**SPARQLWrapper** is a simple Python wrapper around a `SPARQL <https://www.w3.org/TR/sparql11-overview/>`_ service to
remotelly execute your queries. It helps in creating the query
invokation and, possibly, convert the result into a more manageable
format.

"""

__version__ = "2.0.0"
"""The version of SPARQLWrapper"""

__agent__: str = f"sparqlwrapper {__version__} (rdflib.github.io/sparqlwrapper)"


from .SmartWrapper import SPARQLWrapper2
from .sparql_dataframe import get_sparql_dataframe
from .Wrapper import (
    ASK,
    BASIC,
    CONSTRUCT,
    CSV,
    DELETE,
    DESCRIBE,
    DIGEST,
    GET,
    INSERT,
    JSON,
    JSONLD,
    N3,
    POST,
    POSTDIRECTLY,
    RDF,
    RDFXML,
    SELECT,
    TSV,
    TURTLE,
    URLENCODED,
    XML,
    QueryResult,
    SPARQLWrapper,
)

__all__ = [
    "SPARQLWrapper2",
    "get_sparql_dataframe",
    "ASK",
    "BASIC",
    "CONSTRUCT",
    "CSV",
    "DELETE",
    "DESCRIBE",
    "DIGEST",
    "GET",
    "INSERT",
    "JSON",
    "JSONLD",
    "N3",
    "POST",
    "POSTDIRECTLY",
    "RDF",
    "RDFXML",
    "SELECT",
    "TSV",
    "TURTLE",
    "URLENCODED",
    "XML",
    "QueryResult",
    "SPARQLWrapper",
]



#!/usr/bin/python
# -*- coding: utf-8 -*-

from SPARQLWrapper import SPARQLWrapper, JSON, XML, CSV, TSV

sparql = SPARQLWrapper("http://dbpedia.org/sparql")
sparql.setQuery("""
    ASK WHERE { 
        <http://dbpedia.org/resource/Asturias> rdfs:label "Asturias"@es
    }    
""")

# JSON example
print('\n\n*** JSON Example')
sparql.setReturnFormat(JSON)
results = sparql.query().convert()
print(results)

# XML example
print('\n\n*** XML Example')
sparql.setReturnFormat(XML)
results = sparql.query().convert()
print(results.toxml())

# CSV example
print('\n\n*** CSV Example')
sparql.setReturnFormat(CSV)
results = sparql.query().convert()
print(results)

# TSV example
print('\n\n*** TSV Example')
sparql.setReturnFormat(TSV)
results = sparql.query().convert()
print(results)



# -*- coding: utf-8 -*-

"""

..
  Developers involved:

  * Ivan Herman <http://www.ivan-herman.net>
  * Sergio Fernández <http://www.wikier.org>
  * Carlos Tejo Alonso <http://www.dayures.net>
  * Alexey Zakhlestin <https://indeyets.ru/>

  Organizations involved:

  * `World Wide Web Consortium <http://www.w3.org>`_
  * `Foundation CTIC <http://www.fundacionctic.org/>`_

  :license: `W3C® Software notice and license <http://www.w3.org/Consortium/Legal/copyright-software>`_
  :requires: `RDFLib <https://rdflib.readthedocs.io>`_ package.
"""


from typing import Any, Dict, List, Optional, Tuple, Union

from SPARQLWrapper.Wrapper import JSON, SELECT, QueryResult
from SPARQLWrapper.Wrapper import SPARQLWrapper as SW

######################################################################################


class Value(object):
    """
    Class encapsulating a single binding for a variable.

    :ivar variable: The original variable, stored for an easier reference.
    :vartype variable: string
    :ivar value: Value of the binding.
    :vartype value: string
    :ivar type: Type of the binding. One of :attr:`Value.URI`, :attr:`Value.Literal`, :attr:`Value.TypedLiteral`, or
                :attr:`Value.BNODE`.
    :vartype type: string
    :ivar lang: Language tag of the binding, or ``None`` if not set.
    :vartype lang: string
    :ivar datatype: Datatype of the binding, or ``None`` if not set. It is an URI.
    :vartype datatype: string
    """

    URI = "uri"
    """the string denoting a URI variable."""
    Literal = "literal"
    """the string denoting a Literal variable."""
    TypedLiteral = "typed-literal"
    """the string denoting a typed literal variable."""
    BNODE = "bnode"
    """the string denoting a blank node variable."""

    def __init__(self, variable: str, binding: Dict[str, str]) -> None:
        """
        :param variable: the variable for that binding. Stored for an easier reference.
        :type variable: string
        :param binding: the binding dictionary part of the return result for a specific binding.
        :type binding: dict
        """
        self.variable = variable
        self.value = binding["value"]
        self.type = binding["type"]
        self.lang = None
        self.datatype = None
        try:
            self.lang = binding["xml:lang"]
        except:
            # no lang is set
            pass
        try:
            self.datatype = binding["datatype"]
        except:
            pass

    def __repr__(self) -> str:
        cls = self.__class__.__name__
        return "%s(%s:%r)" % (cls, self.type, self.value)


######################################################################################


class Bindings(object):
    """
    Class encapsulating one query result, based on the JSON return format. It decodes the
    return values to make it a bit more usable for a standard usage. The class consumes the
    return value and instantiates a number of attributes that can be consulted directly. See
    the list of variables.

    The `Serializing SPARQL Query Results in JSON <http://www.w3.org/TR/rdf-sparql-json-res/>`_ explains the details of
    the JSON return structures. Very succinctly: the return data has "bindings", which means a list of dictionaries.
    Each dictionary is a possible binding of the SELECT variables to :class:`Value` instances. This structure is made a
    bit more usable by this class.

    :ivar fullResult: The original dictionary of the results, stored for an easier reference.
    :vartype fullResult: dict
    :ivar head: Header part of the return, see the JSON return format document for details.
    :vartype head: dict
    :ivar variables: List of unbounds (variables) of the original query. It is a list of strings. ``None`` in the case
                        of an ASK query.
    :vartype variables: list
    :ivar bindings: The final bindings: list of dictionaries, mapping variables to :class:`Value` instances. \
    If unbound, then no value is set in the dictionary; that can be easily checked with \
    ``var in res.bindings[..]``, for example.
    :vartype bindings: list
    :ivar askResult: by default, set to **False**; in case of an ASK query, the result of the query.
    :vartype askResult: bool
    """

    def __init__(self, retval: QueryResult):
        """
        :param retval: the query result.
        :type retval: :class:`QueryResult<SPARQLWrapper.Wrapper.QueryResult>`
        """
        self.fullResult = retval._convertJSON()
        self.head = self.fullResult["head"]
        self.variables: Optional[List[str]] = None
        try:
            self.variables = self.fullResult["head"]["vars"]
        except:
            pass

        self.bindings: List[Dict[str, Value]] = []
        try:
            for b in self.fullResult["results"]["bindings"]:
                # This is a single binding. It is a dictionary per variable; each value is a dictionary again
                # that has to be converted into a Value instance
                newBind = {}
                # type error: Item "None" of "Union[List[str], Any, None]" has no attribute "__iter__" (not iterable)
                for key in self.variables:  # type: ignore [union-attr]
                    if key in b:
                        # there is a real binding for this key
                        newBind[key] = Value(key, b[key])
                self.bindings.append(newBind)
        except:
            pass

        self.askResult = False
        try:
            self.askResult = self.fullResult["boolean"]
        except:
            pass

    def getValues(self, key: str) -> Optional[List[Value]]:
        """A shorthand for the retrieval of all bindings for a single key. It is
        equivalent to ``[b[key] for b in self[key]]``

        :param key: possible variable name.
        :type key: string
        :return: list of :class:`Value` instances.
        :rtype: list
        """
        try:
            return [b[key] for b in self[key]]
        except:
            return []

    def __contains__(self, key: Union[str, List[str], Tuple[str]]) -> bool:
        """Emulation of the "``key in obj``" operator. Key can be a string for a variable or an array/tuple
        of strings.

        If ``key`` is a variable, the return value is ``True`` if there is at least one binding where ``key`` is
        bound. If ``key`` is an array or tuple, the return value is ``True`` if there is at least one binding
        where *all* variables in ``key`` are bound.

        :param key: possible variable, or array/tuple of variables
        :return: whether there is a binding of the variable in the return
        :rtype: Boolean
        """
        if len(self.bindings) == 0:
            return False
        if type(key) is list or type(key) is tuple:
            # check first whether they are all really variables
            # type error: Unsupported right operand type for in ("Optional[List[str]]")
            if False in [k in self.variables for k in key]:  # type: ignore [operator]
                return False
            for b in self.bindings:
                # try to find a binding where all key elements are present
                if False in [k in b for k in key]:
                    # this is not a binding for the key combination, move on...
                    continue
                else:
                    # yep, this one is good!
                    return True
            return False
        else:
            # type error: Unsupported right operand type for in ("Optional[List[str]]")
            if key not in self.variables:  # type: ignore [operator]
                return False
            for b in self.bindings:
                if key in b:
                    return True
            return False

    def __getitem__(self, key: Union[slice, str, List[str]]) -> List[Dict[str, Value]]:
        """Emulation of the ``obj[key]`` operator.  Slice notation is also available.
        The goal is to choose the right bindings among the available ones. The return values are always
        arrays  of bindings, ie, arrays of dictionaries mapping variable keys to :class:`Value` instances.
        The different value settings mean the followings:

         - ``obj[key]`` returns the bindings where ``key`` has a valid value
         - ``obj[key1,key2,...]`` returns the bindings where *all* ``key1,key2,...`` have valid values
         - ``obj[(key1,key2,...):(nkey1,nkey2,...)]`` returns the bindings where all ``key1,key2,...`` have
         valid values and *none* of the ``nkey1,nkey2,...`` have valid values
         - ``obj[:(nkey1,nkey2,...)]`` returns the bindings where *none* of the ``nkey1,nkey2,...`` have valid values

        In all cases complete bindings are returned, ie, the values for other variables, not present among
        the keys in the call, may or may not be present depending on the query results.

        :param key: possible variable or array/tuple of keys with possible slice notation
        :return: list of bindings
        :rtype: array of variable -> :class:`Value`  dictionaries
        """

        def _checkKeys(keys: Union[List[Any], Tuple[Any, ...]]) -> bool:
            if len(keys) == 0:
                return False
            for k in keys:
                # type error: Unsupported right operand type for in ("Optional[List[str]]")
                if (
                    not isinstance(k, str)
                    or k not in self.variables  # type: ignore [operator]
                ):
                    return False
            return True

        def _nonSliceCase(
            key: Union[
                str,
                List[Any],
                Tuple[Any],
            ]
        ) -> Union[List[Any], bool, Tuple[Any]]:
            # type error: Unsupported right operand type for in ("Optional[List[str]]")
            if isinstance(key, str) and key != "" and key in self.variables:  # type: ignore[operator]
                # unicode or string:
                return [key]
            elif type(key) is list or type(key) is tuple:
                if _checkKeys(key):
                    return key
            return False

        # The arguments should be reduced to arrays of variables, ie, unicode strings
        yes_keys: Union[List[Any], bool, Tuple[Any]] = []
        no_keys: Union[List[Any], bool, Tuple[Any]] = []
        if type(key) is slice:
            # Note: None for start or stop is all right
            if key.start:
                yes_keys = _nonSliceCase(key.start)
                if not yes_keys:
                    raise TypeError
            if key.stop:
                no_keys = _nonSliceCase(key.stop)
                if not no_keys:
                    raise TypeError
        else:
            yes_keys = _nonSliceCase(key)

        # got it right, now get the right binding line with the constraints
        retval: List[Dict[str, Value]] = []
        for b in self.bindings:
            # first check whether the 'yes' part is all there:
            # type error: Item "bool" of "Union[List[Any], bool, Tuple[Any]]" has no attribute "__iter__" (not iterable)
            if False in [k in b for k in yes_keys]:  # type: ignore[union-attr]
                continue
            # type error: Item "bool" of "Union[List[Any], bool, Tuple[Any]]" has no attribute "__iter__" (not iterable)
            if True in [k in b for k in no_keys]:  # type: ignore[union-attr]
                continue
            # if we got that far, we should be all right!
            retval.append(b)
        # if retval is of zero length, no hit; an exception should be raised to stay within the python style
        if len(retval) == 0:
            raise IndexError
        return retval

    def convert(self) -> "Bindings":
        """This is just a convenience method, returns ``self``.

        Although :class:`SPARQLWrapper2.Bindings` is not a subclass of
        :class:`SPARQLWrapper.QueryResult<SPARQLWrapper.Wrapper.QueryResult>`, it is returned as a result by
        :func:`SPARQLWrapper2.query`, just like :class:`QueryResult<SPARQLWrapper.Wrapper.QueryResult>` is returned by
        :func:`SPARQLWrapper.query()<SPARQLWrapper.Wrapper.SPARQLWrapper.query>`. Consequently,
        having an empty :func:`convert` method to imitate
        :class:`QueryResult's convert() method<SPARQLWrapper.Wrapper.QueryResult.convert>`
        may avoid unnecessary problems.
        """
        return self


##############################################################################################################


class SPARQLWrapper2(SW):
    """Subclass of :class:`~SPARQLWrapper.Wrapper.SPARQLWrapper` that works with a JSON SELECT return result only. The
    query result is automatically set to a :class:`Bindings` instance. Makes the average query processing a bit
    simpler..."""

    def __init__(self, baseURI: str, defaultGraph: Optional[str] = None):
        """
        Class encapsulating a full SPARQL call. In contrast to the :class:`~SPARQLWrapper.Wrapper.SPARQLWrapper`
        superclass, the return format cannot be set (it is defaulted to
        :attr:`~SPARQLWrapper.Wrapper.SPARQLWrapper.JSON`).

        :param baseURI: string of the SPARQL endpoint's URI.
        :type baseURI: string
        :param defaultGraph: URI for the default graph. Default is ``None``, can be set via an explicit call, too.
        :type defaultGraph: string
        """
        super(SPARQLWrapper2, self).__init__(
            baseURI, returnFormat=JSON, defaultGraph=defaultGraph
        )

    def setReturnFormat(self, format: Optional[str]) -> None:
        """
        Set the return format (:meth:`overriding the inherited method
        <SPARQLWrapper.Wrapper.SPARQLWrapper.setReturnFormat>`).

        .. warning::

           This method does nothing; this class instance should work with JSON only. The method is defined \
           just to avoid possible errors by erroneously setting the return format. \
           When using this class, the user can safely ignore this call.

        :param format: return format
        :type format: string
        """
        pass

    def query(self) -> Union[Bindings, QueryResult]:  # type: ignore[override]
        """
        Execute the query and do an automatic conversion.

        Exceptions can be raised if either the URI is wrong or the HTTP sends back an error.
        The usual urllib2 exceptions are raised, which cover possible SPARQL errors, too.

        If the query type is *not* SELECT, the method falls back to the
        :meth:`corresponding method in the superclass<SPARQLWrapper.Wrapper.SPARQLWrapper.query>`.

        :return: query result
        :rtype: :class:`Bindings` instance
        """
        res = super(SPARQLWrapper2, self).query()

        if self.queryType == SELECT:
            return Bindings(res)
        else:
            return res

    def queryAndConvert(  # type: ignore[override]
        self,
    ) -> Union[Union[Bindings, QueryResult], QueryResult.ConvertResult]:
        """This is here to override the inherited method; it is equivalent to :class:`query`.

        If the query type is *not* SELECT, the method falls back to the
        :meth:`corresponding method in the superclass<SPARQLWrapper.Wrapper.SPARQLWrapper.queryAndConvert>`.

        :return: the converted query result.
        """
        if self.queryType == SELECT:
            return self.query()
        else:
            return super(SPARQLWrapper2, self).queryAndConvert()



# -*- coding: utf-8 -*-

"""
SPARQL Wrapper exceptions

..
  Developers involved:

  * Ivan Herman <http://www.ivan-herman.net>
  * Sergio Fernández <http://www.wikier.org>
  * Carlos Tejo Alonso <http://www.dayures.net>
  * Alexey Zakhlestin <https://indeyets.ru/>

  Organizations involved:

  * `World Wide Web Consortium <http://www.w3.org>`_
  * `Foundation CTIC <http://www.fundacionctic.org/>`_

  :license: `W3C® Software notice and license <http://www.w3.org/Consortium/Legal/copyright-software>`_

"""

from typing import Optional


class SPARQLWrapperException(Exception):
    """
    Base class for SPARQL Wrapper exceptions
    """

    msg = "an exception has occurred"

    def __init__(self, response: Optional[bytes] = None):
        """
        :param string response: The server response
        """
        if response:
            formatted_msg = "%s: %s. \n\nResponse:\n%r" % (
                self.__class__.__name__,
                self.msg,
                response,
            )
        else:
            formatted_msg = "%s: %s." % (self.__class__.__name__, self.msg)

        super(SPARQLWrapperException, self).__init__(formatted_msg)


class EndPointInternalError(SPARQLWrapperException):
    """
    Exception type for Internal Server Error responses. Usually HTTP response status code ``500``.
    """

    msg = "The endpoint returned the HTTP status code 500"


class QueryBadFormed(SPARQLWrapperException):
    """
    Query Bad Formed exception. Usually HTTP response status code ``400``.
    """

    msg = "A bad request has been sent to the endpoint: probably the SPARQL query is badly formed"


class EndPointNotFound(SPARQLWrapperException):
    """
    End Point Not Found exception. Usually HTTP response status code ``404``.
    """

    msg = "It was not possible to connect to the given endpoint: check it is correct"


class Unauthorized(SPARQLWrapperException):
    """
    Access is denied due to invalid credentials (unauthorized). Usually HTTP response status code ``401``.

    .. versionadded:: 1.8.2
    """

    msg = "Access to that endpoint is denied due to invalid credentials (unauthorized). Check the credentials"


class URITooLong(SPARQLWrapperException):
    """
    The URI requested by the client is longer than the server is willing to interpret. Usually HTTP response
    status code ``414``.

    .. versionadded:: 1.8.3
    """

    msg = (
        "The URI requested by the client is longer than the server is willing to interpret. "
        "Check if the request was sent using GET method instead of POST method."
    )



"""
Query a SPARQL endpoint and return results as a Pandas dataframe.
"""
import io
from typing import TYPE_CHECKING, Any, Dict, List, Union

from SPARQLWrapper.SmartWrapper import Bindings, SPARQLWrapper2, Value
from SPARQLWrapper.Wrapper import CSV, SELECT, SPARQLWrapper

if TYPE_CHECKING:
    import pandas as pd


class QueryException(Exception):
    pass


def get_sparql_dataframe_orig(
    endpoint: str, query: Union[str, bytes]
) -> "pd.DataFrame":
    """copy paste from: https://github.com/lawlesst/sparql-dataframe"""
    # pandas inside to avoid requiring it
    import pandas as pd

    sparql = SPARQLWrapper(endpoint)
    sparql.setQuery(query)
    if sparql.queryType != SELECT:
        raise QueryException("Only SPARQL SELECT queries are supported.")
    sparql.setReturnFormat(CSV)
    results = sparql.query().convert()
    if isinstance(results, bytes):
        _csv = io.StringIO(results.decode("utf-8"))
        return pd.read_csv(_csv, sep=",")
    else:
        raise TypeError(type(results))


def get_sparql_typed_dict(
    endpoint: str, query: Union[str, bytes]
) -> List[Dict[str, Value]]:
    """modified from: https://github.com/lawlesst/sparql-dataframe"""
    # pandas inside to avoid requiring it
    import pandas as pd
    # rdflib in here because there is some meta stuff in the setup.py and Travis fails because rdflib is installed later
    import rdflib.term
    sparql = SPARQLWrapper2(endpoint)
    sparql.setQuery(query)
    if sparql.queryType != SELECT:
        raise QueryException("Only SPARQL SELECT queries are supported.")
    # sparql.setReturnFormat(JSON)
    results = sparql.query()
    if not isinstance(results, Bindings):
        raise TypeError(type(results))
    # consider perf hacking later, probably slow
    # convert list of dicts to python types
    d = []
    for x in results.bindings:
        row = {}
        for k in x:
            v = x[k]
            vv = rdflib.term.Literal(v.value, datatype=v.datatype).toPython()  # type: ignore[no-untyped-call]
            row[k] = vv
        d.append(row)
    return d


def get_sparql_dataframe(endpoint: str, query: Union[str, bytes]) -> "pd.DataFrame":
    # pandas inside to avoid requiring it
    import pandas as pd

    d = get_sparql_typed_dict(endpoint, query)
    # TODO: will nan fill somehow, make more strict if there is way of getting the nan types from rdflib
    df = pd.DataFrame(d)
    return df



#!/usr/bin/env python

# -*- coding: utf-8 -*-

import argparse
import json
import os
import shutil
import sys
import xml
from typing import List, Optional

import rdflib

from . import __version__
from .Wrapper import SPARQLWrapper, _allowedAuth, _allowedFormats, _allowedRequests


class SPARQLWrapperFormatter(
    argparse.ArgumentDefaultsHelpFormatter, argparse.RawDescriptionHelpFormatter
):
    pass


def check_file(v: str) -> str:
    if os.path.isfile(v):
        return v
    elif v == "-":
        return "-"  # stdin
    else:
        raise argparse.ArgumentTypeError("file '%s' is not found" % v)


def choicesDescriptions() -> str:
    d = "\n  - ".join(["allowed FORMAT:"] + _allowedFormats)
    d += "\n  - ".join(["\n\nallowed METHOD:"] + _allowedRequests)
    d += "\n  - ".join(["\n\nallowed AUTH:"] + _allowedAuth)
    return d


def parse_args(test: Optional[List[str]] = None) -> argparse.Namespace:
    """Parse arguments."""
    parser = argparse.ArgumentParser(
        prog="rqw",
        formatter_class=(
            lambda prog: SPARQLWrapperFormatter(
                prog,
                **{
                    "width": shutil.get_terminal_size(fallback=(120, 50)).columns,
                    "max_help_position": 30,
                },
            )
        ),
        description="sparqlwrapper CLI",
        epilog=choicesDescriptions(),
    )
    input_group = parser.add_mutually_exclusive_group(required=True)
    input_group.add_argument(
        "-f",
        "--file",
        metavar="FILE",
        type=check_file,
        help="query with sparql file (stdin: -)",
    )
    input_group.add_argument("-Q", "--query", metavar="QUERY", help="query with string")
    parser.add_argument(
        "-F",
        "--format",
        default="json",
        metavar="FORMAT",
        choices=_allowedFormats,
        help="response format",
    )
    parser.add_argument(
        "-e",
        "--endpoint",
        metavar="URI",
        help="sparql endpoint",
        default="http://dbpedia.org/sparql",
    )
    parser.add_argument(
        "-m",
        "--method",
        metavar="METHOD",
        choices=_allowedRequests,
        help="request method",
    )
    parser.add_argument(
        "-a", "--auth", metavar="AUTH", choices=_allowedAuth, help="HTTP auth"
    )
    parser.add_argument(
        "-u", "--username", metavar="ID", default="guest", help="username for auth"
    )
    parser.add_argument(
        "-p", "--password", metavar="PW", default="", help="password for auth"
    )
    parser.add_argument("-q", "--quiet", action="store_true", help="supress warnings")
    parser.add_argument(
        "-V", "--version", action="version", version="%(prog)s {}".format(__version__)
    )
    if test is None:
        return parser.parse_args()
    else:
        return parser.parse_args(test)


def main(test: Optional[List[str]] = None) -> None:
    args = parse_args(test)
    if args.quiet:
        import warnings

        warnings.filterwarnings("ignore")

    q = ""
    if args.query is not None:
        q = args.query
    elif args.file == "-":
        q = sys.stdin.read()
    else:
        q = open(args.file, "r").read()

    sparql = SPARQLWrapper(
        args.endpoint,
        agent=(
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) "
            "AppleWebKit/537.36 (KHTML, like Gecko) "
            "Chrome/96.0.4664.110 Safari/537.36"
        ),
    )
    if args.auth is not None:
        sparql.setHTTPAuth(args.auth)
        sparql.setCredentials(args.username, args.password)
    if args.method is not None:
        sparql.setMethod(args.method)
    sparql.setQuery(q)
    sparql.setReturnFormat(args.format)
    results = sparql.query().convert()

    if isinstance(results, dict):
        # "json"
        print(json.dumps(results, indent=4))
    elif isinstance(results, xml.dom.minidom.Document):
        # "xml"
        print(results.toxml())
    elif isinstance(results, bytes):
        # "csv", "tsv", "turtle", "n3"
        print(results.decode("utf-8"))
    elif isinstance(results, rdflib.graph.ConjunctiveGraph):
        # "rdf"
        print(results.serialize())
    else:
        # unknown type
        raise TypeError(f"Unsupported result of type {type(results)}: {results!r}")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""
A simple implementation of a key case-insensitive dictionary.
..
  Developers involved:
  * Ivan Herman <http://www.ivan-herman.net>
  * Sergio Fernández <http://www.wikier.org>
  * Carlos Tejo Alonso <http://www.dayures.net>
  * Alexey Zakhlestin <https://indeyets.ru/>
  Organizations involved:
  * `World Wide Web Consortium <http://www.w3.org>`_
  * `Foundation CTIC <http://www.fundacionctic.org/>`_
  :license: `W3C® Software notice and license <http://www.w3.org/Consortium/Legal/copyright-software>`_
"""

from typing import Dict, Mapping, TypeVar

_V = TypeVar("_V")

class KeyCaseInsensitiveDict(Dict[str, _V]):
    """
    A simple implementation of a key case-insensitive dictionary
    """

    def __init__(self, d: Mapping[str, _V]={}) -> None:
        """
        :param dict d: The source dictionary.
        """
        for k, v in d.items():
            self[k] = v

    def __setitem__(self, key: str, value: _V) -> None:
        if hasattr(key, "lower"):
            key = key.lower()
        dict.__setitem__(self, key, value)

    def __getitem__(self, key: str) -> _V:
        if hasattr(key, "lower"):
            key = key.lower()
        return dict.__getitem__(self, key)

    def __delitem__(self, key: str) -> None:
        if hasattr(key, "lower"):
            key = key.lower()
        dict.__delitem__(self, key)



from SPARQLWrapper import SPARQLWrapper, RDFXML, N3, TURTLE, JSONLD
from rdflib import Graph

sparql = SPARQLWrapper("http://dbpedia.org/sparql")

sparql.setQuery("""
    DESCRIBE <http://dbpedia.org/resource/Asturias>
""")

# RDF/XML example
print('\n\n*** RDF/XML Example')
sparql.setReturnFormat(RDFXML)
results = sparql.query().convert()
print(results.serialize(format='xml'))

# N3 example
print('\n\n*** N3 Example')
sparql.setReturnFormat(N3)
results = sparql.query().convert()
g = Graph()
g.parse(data=results, format="n3")
print(g.serialize(format='n3'))

# Turtle example
print('\n\n*** TURTLE Example')
sparql.setReturnFormat(TURTLE)
results = sparql.query().convert()
g = Graph()
g.parse(data=results, format="turtle")
print(g.serialize(format='turtle'))

# JSONLD example
print('\n\n*** JSONLD Example')
sparql.setReturnFormat(JSONLD)
results = sparql.query().convert()
print(results.serialize(format='json-ld'))



#!/usr/bin/python
# -*- coding: utf-8 -*-

from SPARQLWrapper import SPARQLWrapper, JSON

sparql = SPARQLWrapper("http://agrovoc.uniroma2.it:3030/agrovoc/sparql")

query = """
PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>

SELECT ?conceptAGROVOC ?conceptGEMET ?label
WHERE {
    ?conceptAGROVOC rdf:type skos:Concept ;
                    skos:prefLabel ?label;
                    skos:inScheme <http://voc.landportal.info/landterms> .
    FILTER (lang(?label) = 'en')

    SERVICE <http://semantic.eea.europa.eu/sparql> { 
            ?conceptGEMET skos:prefLabel ?label ; 
            skos:inScheme <http://www.eionet.europa.eu/gemet/gemetThesaurus> .
    } 
}
"""
sparql.setQuery(query)
sparql.setReturnFormat(JSON)

results = sparql.query().print_results()



#!/usr/bin/python
# -*- coding: utf-8 -*-

import sys
import getopt
from SPARQLWrapper import SPARQLWrapper2

localSparqler = "http://localhost:2020/sparql"
localVirtuoso = "http://localhost:8890/sparql"

def main(server, query, sponge=False):
    sparql = SPARQLWrapper2(server)
    if sponge:
        sparql.addExtraURITag("should-sponge", "grab-everything")

    sparql.setQuery(query)
    res = sparql.query()
    variables = res.variables

    print("Variables:")
    print(variables)
    print()
    print("Bindings:")
    for b in res.bindings:
        for v in res.variables:
            try:
                val = b[v]
                if val.lang:
                    str = "%s: %s@%s" % (v, val.value, val.lang)
                elif val.datatype:
                    str = "%s: %s^^%s" % (v, val.value, val.datatype)
                else:
                    str = "%s: %s" % (v, val.value)
            except KeyError:
                # no binding to that one...
                str = "%s: <<None>>" % v
            print(str.encode('utf-8'))
        print()



# -------------------------------------------------------------------------------------------------------------
server = localSparqler
query = ""
sponge = False
usagetxt = """%s [-s] [-u url] [file]
-s:      use local sparqler (default)
-v:      use local virtuoso
-u url:  server url
-p:      issue an extra sponge for virtuoso
file: sparql query file
"""
def usage():
    print(usagetxt % sys.argv[0])
    sys.exit(1)

if __name__ == '__main__':
    if len(sys.argv) == 1:
        usage()
    try:
        opts, args = getopt.getopt(sys.argv[1:], "shu:pv")
        for o, a in opts:
            if o == "-s":
                server = localSparqler
            elif o == "-v":
                server = localVirtuoso
                sponge = True
            elif o == "-h":
                print(usage)
                sys.exit(0)
            elif o == "-u":
                server = a
            elif o == "-p":
                sponge = True
        if query == "" and len(args) > 0:
            inp = file(args[0])
            query = ""
            for l in inp:
                query += l
    except:
        usage()
    if query == "":
        usage()
    main(server, query, sponge)
